package com.vividsolutions.jump.workbench.ui.plugin.cormat;

import java.io.*;

import java.io.IOException;
import java.util.regex.*;

import com.vividsolutions.jts.geom.Coordinate;

/** 
 * This class represents the CorMat Data (Correction Matrix).
 * <code>CormatData</code> A set of points distributed on an arrayed grid with the 
 * interval of 100m that covers the entire corresponding 
 * NTDB (National Topographic Data Base) data set.
 */
public class CormatData {

	private static int maxCormatxIndex = 500; // Maximum width = 50000m
	private static int maxCormatyIndex = 500; // Maximum height = 50000m
	
	private double minEasting;  // Minimum Easting
	private double maxEasting;  // Maximum Easting
	private double minNorthing; // Minimum Northing;
	private double maxNorthing; // Maximum Northing;
	
	private boolean wasOutlier = false;  // Was outlier flag 

    /**
	 *  The CormatData Matrix.
	 */
	public CormatDataPoint[][] CormatDataMatrix; // CorMat Data Matrix

    /**
     *  Constructs an default initialized <code>CormatData</code>.
     */    
    public CormatData(){
    	// Constructor.
    	minEasting = (double)0.0;
    	maxEasting = (double)0.0;
    	minNorthing = (double)0.0;
    	maxNorthing = (double)0.0;
    	CormatDataMatrix = new CormatDataPoint[maxCormatxIndex][maxCormatyIndex];
    };

    /**
     *  To return the <code>CormatDataPoint</data> for this index (i,j) into the
     *  CorMat Data Matrix
     *
     *@param  i  into CorMat Data Matrix
     *@param  j  into CorMat Data Matrix
     *@return <code>CormatDataPoint</code> from CormatDataMatrix(i, j)
     */
    public CormatDataPoint getDataPointbyIndex(int i, int j){
    	// Check if index in bounds
    	if (indexInBounds(i, j) != 0) return(new CormatDataPoint(0.0,0.0,0.0,0.0));
       return CormatDataMatrix[i][j];	
    }

    /**
     * Check if there were any requests for a location outside the loaded CorMat Matrix
     *  
     *@return <code>true</code> if there was an out of bound index request.
     */
    public boolean getWasOutlier(){
    	return(wasOutlier);
    }
    
    /**
     *  To load the specified cormat file into the CorMat Data Matrix
     *
     *@param  cormatFile  
     *@return 0 if OK; 1 otherwise
     *@throws IOException
     */
    public int LoadCormatFile(File cormatFile) throws IOException {
    	
		try {
			BufferedReader in = new BufferedReader(new FileReader(cormatFile.getPath()));
			String str;
			String patternStr = " ";
			String[] strflds = new String[4];
			int pn = 0; // Point Number
			while ((str = in.readLine()) != null)
			{
				
				// Increment Point Number
				pn = pn + 1;
				
				str = removeDuplicateWhitespace(str);
				
				strflds = str.split(patternStr);
				CormatDataPoint cdp = new CormatDataPoint();
				cdp.easting = Double.parseDouble(strflds[0]);
				cdp.northing = Double.parseDouble(strflds[1]);
				cdp.deltax = Double.parseDouble(strflds[2]);
				cdp.deltay = Double.parseDouble(strflds[3]);
				
				// Check if First Point (ie: SouthWest Corner)
				if (pn == 1) {
					// Set Minimum Easting and Minimum Northing
					// Used for indexing into the CormatDataMatrix 
					minEasting = cdp.easting;
					minNorthing = cdp.northing;
					// Set Maximum Easting and Maximum Northing
					// Used for indexing into the CormatDataMatrix 
					maxEasting = cdp.easting;
					maxNorthing = cdp.northing;
				}
				else
				{
					// Verify that minEasting and minNorthing
					// are the real mins
					// And look for maximum easting and northing
					if (cdp.easting < minEasting)
					{
//						System.out.println("WOLF: Error found a new Min Easting at point " + pn);
						return(1);
					}
					else if (cdp.easting > maxEasting)
					{
						maxEasting = cdp.easting;
					}; /*if*/
					if (cdp.northing < minNorthing)
					{
//						System.out.println("WOLF: Error found a new Min Northing at point " + pn);
						return(1);
					}
					else if (cdp.northing > maxNorthing)
					{
						maxNorthing = cdp.northing;
					}; /*if*/
				}; /*if*/
		
				/* Add Cormat Data Point */
				if (addByCoordinate(cdp) != 0)
				{
				};
				
			}; /*while*/
		}
		catch (IOException e)
		{
			throw e;
		};
    	
		// Return OK
    	return(0);
    	
    }

    /**
     * Check if index (i,j) is in bounds of the CorMat Data Matrix
     *  
     *@param  i  index into CorMat Data Matrix
     *@param  j  index into CorMat Data Matrix
     *@return 0 if within bounds; 1 otherwise
     */
    public int indexInBounds(int i, int j){
    	// Check if index in bounds
    	if ((i <0) || (j < 0) || (i >= maxCormatxIndex) || (j >= maxCormatyIndex))
    	{
    		// Set Was Outlier Flag 
    		wasOutlier = true;
    		// Error: Index out of bounds
    		return(1);
    	}; /*if*/
    	// Return Index in bounds
        return(0);	
    }
    
    /**
     * Add a CormatDataPoint at index (i,j) into the CorMat Data Matrix
     *  
     *@param  cdp <data>CormatDataPoint</data> to add.
     *@param  i  index into CorMat Data Matrix
     *@param  j  index into CorMat Data Matrix
     *@return 0 if within bounds; 1 otherwise
     */
    public int addByIndex(CormatDataPoint cdp, int i, int j){
    	// Check if index in bounds
    	if (indexInBounds(i, j) != 0) return(1); // Index out of bounds
    	// Assign to the CorMat Data Matrix
    	CormatDataMatrix[i][j] = cdp;	
    	// Return Index in bounds
    	return(0);
    }

    /**
     * Get a CormatDataPoint at index (i,j) from the CorMat Data Matrix
     *  
     *@param  i  index into CorMat Data Matrix
     *@param  j  index into CorMat Data Matrix
     *@return cdp <data>CormatDataPoint</data> from CorMat Data Matrix at (i,j) 
     */
    public CormatDataPoint getByIndex(int i, int j){
    	// Check if index in bounds
    	if (indexInBounds(i, j) != 0) return(new CormatDataPoint(0.0,0.0,0.0,0.0));
    	return(CormatDataMatrix[i][j]);	
    }

    /**
     * Add a <data>CormatDataPoint</data> by it's location (easting, northing) 
     * into the CorMat Data Matrix
     *  
     *@param  cdp <data>CormatDataPoint</data> to add.
     *@return 0 if within bounds; 1 otherwise
     */
    public int addByCoordinate(CormatDataPoint cdp){
    	double dxi;
    	double dyi;
    	int xi; // x-index;
    	int yi; // y-index;
    	dxi = (cdp.easting - minEasting) / 100;
    	dyi = (cdp.northing - minNorthing) / 100;
    	xi = (int)dxi; 
    	yi = (int)dyi;
    	// Check if index in bounds
    	if (indexInBounds(xi, yi) != 0) return(1);
    	CormatDataMatrix[xi][yi] = cdp;
    	return(0);
    }

    /**
     * Get a CormatDataPoint for location (easting, northing) from the CorMat Data Matrix
     *  
     *@param  e   easting  
     *@param  n   northing  
     *@return cdp <data>CormatDataPoint</data> from CorMat Data Matrix at (i,j) 
     */
    public CormatDataPoint getByEastingAndNorthing(double e, double n){
    	double dxi;
    	double dyi;
    	int xi; // x-index;
    	int yi; // y-index;
    	dxi = (e - minEasting) / 100;
    	dyi = (n - minNorthing) / 100;
    	xi = (int)dxi; 
    	yi = (int)dyi;
    	// Check if index in bounds
    	if (indexInBounds(xi, yi) != 0) return(new CormatDataPoint(0.0,0.0,0.0,0.0));
    	return(CormatDataMatrix[xi][yi]);	
    }

    /* from http://www.cits.rncan.gc.ca/cit/servlet/CIT/site_id=01&page_id=1-005-002-010-002-001.html
    APPENDIX A: Bilinear Interpolation Algorithm for Planimetric Correction
    Definitions of the Mnemonics Used
    XMAX	Maximum X coordinate in the correction matrix
    XMIN	Minimum X coordinate in the correction matrix
    SQRT	Square root function
    R100	Function rounded off to the nearest 100 units (100=R100(149.5), 200=R100(150.0))
    READ	Read function


    For each coordinate (X, Y) in the original NTDB data set...;
    Define the coordinates of the four adjacent points in the matrix;
       CX[1]=R100(X-49.5); CX[2]=R100(CX[1]);  CX[3]=R100(X+49.5); CX[4]=R100(CX[3]);
       CY[1]=R100(Y+49.5); CY[2]=R100(Y-49.5); CY[3]=R100(CY[2]);  CY[4]=R100(CY[1]);

       Processing loop for the four adjacent points "P" (P=1 to 4);
          Find the line "LINE" which contains the point "P" in the matrix;
          LINE=1 + (CX[P]-XMIN)/100 + (CY[P]-YMIN)/(XMAX-XMIN+100);

          Read the point "P" from the line "LINE" in the correction matrix;
          READ(CX[P] CY[P] DX[P] DY[P]);

          Calculate the weighting factor associated with the point "P";
          WG[P]=1-(SQRT((X-CX[P])**2 + (Y-CY[P])**2)/SQRT(100**2 + 100**2));
       End loop;

       Interpolation of the correction to be made;
       SUMWX=DX[1]*WG[1] + DX[2]*WG[2] + DX[3]*WG[3] + DX[4]*WG[4];
       SUMWY=DY[1]*WG[1] + DY[2]*WG[2] + DY[3]*WG[3] + DY[4]*WG[4];

       SUMW= WG[1] + WG[2] + WG[3] + WG[4];

       DXBL=SUMWX/SUMW;
       DYBL=SUMWY/SUMW;

       Planimetric correction of the coordinates with the interpolated value;
       X=X+DXBL;  Y=Y+DYBL;
*/
    /**
     * To apply a CorMat matrix offset for coordinate oc
     *  
     *@param  oc <data>Coordinate</data> Old Coordinate  
     *@return nc <data>Coordinate</data> New Coordinate (Old Coordinate + CorMat Offset) 
     */
    public Coordinate cormatOffset(Coordinate oc){
    	
    	CormatDataPoint[] adp = new CormatDataPoint[4]; // 4 Adjacent points (ie: the bounding box points)
    	double[] wg = new double[4];  					// Weighting
    	double sumwx;									// Sum of the Interpolated X Weightings
    	double sumwy;									// Sum of the Interpolated Y Weightings
    	double sumw;									// Sum of the Weightings
    	double dxbl;									// Interpolated Delta-X
    	double dybl;									// Interpolated Delta-Y
    	
//    	adp[0] = getByEastingAndNorthing(oc.x - 49.5, oc.y + 49.5);
//    	adp[1] = getByEastingAndNorthing(adp[0].easting, oc.y - 49.5);
//    	adp[2] = getByEastingAndNorthing(oc.x + 49.5, adp[1].northing);
//    	adp[3] = getByEastingAndNorthing(adp[2].easting, adp[0].northing);
   	
    	// Above follows the original algorithm; however, instead of doing the rounding later
    	// applied it here before it's get truncated in getByEastingAndNorthing
    	// Also; by using getByEastingandNorthing we have the value indexed into the CormatDataMatrix
    	adp[0] = getByEastingAndNorthing(oc.x, oc.y + 100.0);
    	adp[1] = getByEastingAndNorthing(adp[0].easting, oc.y);
    	adp[2] = getByEastingAndNorthing(oc.x + 100.0, adp[1].northing);
    	adp[3] = getByEastingAndNorthing(adp[2].easting, adp[0].northing);
    	// Calculate Weightings
    	int i;
    	for (i = 0; i < 4; i++)
    	{
    	   wg[i] = 1 - (Math.sqrt(((oc.x - adp[i].easting) * (oc.x - adp[i].easting)) + ((oc.y - adp[i].northing) * (oc.y - adp[i].northing))) 
    			   			/ Math.sqrt(10000.0 + 10000.0));
    	}; /*for*/
    	// Interpolation of the correcion to be made
    	sumwx = (adp[0].deltax * wg[0]) + (adp[1].deltax * wg[1]) + (adp[2].deltax * wg[2]) + (adp[3].deltax * wg[3]); 
    	sumwy = (adp[0].deltay * wg[0]) + (adp[1].deltay * wg[1]) + (adp[2].deltay * wg[2]) + (adp[3].deltay * wg[3]);
    	
    	sumw = wg[0] + wg[1] + wg[2] + wg[3];
    	
    	dxbl = sumwx / sumw;
    	dybl = sumwy / sumw;
    		
    	// Planimetric correction of the coordinates with the interpolated value
    	Coordinate nc = new Coordinate();
    	
    	// Determine new coordinate
    	nc.x = oc.x + dxbl;
    	nc.y = oc.y + dybl;
    	
    	return(nc);
    }
    
    // Function to remove duplicate white space characters from a sting
    private static String removeDuplicateWhitespace(String inputStr) {
        String patternStr = "\\s+";
        String replaceStr = " ";
        Pattern pattern = Pattern.compile(patternStr);
        Matcher matcher = pattern.matcher(inputStr);
        return matcher.replaceAll(replaceStr);
    }
    
}

